@@ -1,16 +1,22 @@
+/*
+ * This Source Code Form is subject to the terms of the Mozilla Public
+ * License, v. 2.0. If a copy of the MPL was not distributed with this
+ * file, You can obtain one at https://mozilla.org/MPL/2.0/.
+ */
+
 use crate::{
-    CameraFormat, CameraIndexType, CameraInfo, CaptureAPIBackend, CaptureBackendTrait, FrameFormat,
-    NokhwaError, Resolution,
+    CameraControl, CameraFormat, CameraIndexType, CameraInfo, CaptureAPIBackend,
+    CaptureBackendTrait, FrameFormat, KnownCameraControls, NokhwaError, Resolution,
 };
 use image::{ImageBuffer, Rgb};
 use opencv::{
-    core::{Mat, MatTrait, MatTraitManual, Vec3b},
+    core::{Mat, MatTraitConst, MatTraitConstManual, Vec3b},
     videoio::{
-        VideoCapture, VideoCaptureTrait, CAP_ANY, CAP_AVFOUNDATION, CAP_MSMF, CAP_PROP_FPS,
-        CAP_PROP_FRAME_HEIGHT, CAP_PROP_FRAME_WIDTH, CAP_V4L2,
+        VideoCapture, VideoCaptureTrait, VideoCaptureTraitConst, CAP_ANY, CAP_AVFOUNDATION,
+        CAP_MSMF, CAP_PROP_FPS, CAP_PROP_FRAME_HEIGHT, CAP_PROP_FRAME_WIDTH, CAP_V4L2,
     },
 };
-use std::collections::HashMap;
+use std::{any::Any, borrow::Cow, collections::HashMap};
 
 /// Converts $from into $to
 /// Example usage:
@@ -45,11 +51,14 @@ macro_rules! tryinto_num {
 ///  - This backend can also do IP Camera input.
 ///  - The backend's backend will default to system level APIs on Linux(V4L2), Mac(AVFoundation), and Windows(Media Foundation). Otherwise, it will decide for itself.
 ///  - If the [`OpenCvCaptureDevice`] is initialized as a `IPCamera`, the [`CameraFormat`]'s `index` value will be [`u32::MAX`](std::u32::MAX) (4294967295).
-///  - `OpenCV` does not support camera querying. Camera Name and Camera supported resolution/fps/fourcc is a [`UnsupportedOperation`](NokhwaError::UnsupportedOperation).
+///  - `OpenCV` does not support camera querying. Camera Name and Camera supported resolution/fps/fourcc is a [`UnsupportedOperationError`](NokhwaError::UnsupportedOperationError).
 /// Note: [`resolution()`](crate::CaptureBackendTrait::resolution()), [`frame_format()`](crate::CaptureBackendTrait::frame_format()), and [`frame_rate()`](crate::CaptureBackendTrait::frame_rate()) is not affected.
 ///  - [`CameraInfo`]'s human name will be "`OpenCV` Capture Device {location}"
 ///  - [`CameraInfo`]'s description will contain the Camera's Index or IP.
 ///  - The API Preference order is the native OS API (linux => `v4l2`, mac => `AVFoundation`, windows => `MSMF`) than [`CAP_AUTO`](https://docs.opencv.org/4.5.2/d4/d15/group__videoio__flags__base.html#gga023786be1ee68a9105bf2e48c700294da77ab1fe260fd182f8ec7655fab27a31d)
+/// - The `Any` type for [`raw_camera_control()`](CaptureBackendTrait::raw_camera_control) is [`i32`], and its return `Any` is a [`f64`]. Please check [`OpenCV Documentation Constants`](https://docs.rs/opencv/0.53.1/opencv/videoio/index.html) for more.
+/// - The `Any` type for `control` for [`set_raw_camera_control()`](CaptureBackendTrait::set_raw_camera_control) is [`i32`] and [`f64`]. Please check [`OpenCV Documentation Constants`](https://docs.rs/opencv/0.53.1/opencv/videoio/index.html) for more.
+
 pub struct OpenCvCaptureDevice {
     camera_format: CameraFormat,
     camera_location: CameraIndexType,
@@ -100,13 +109,18 @@ impl OpenCvCaptureDevice {
                         index = idx;
                         vc
                     }
-                    Err(why) => return Err(NokhwaError::CouldntOpenDevice(why.to_string())),
+                    Err(why) => {
+                        return Err(NokhwaError::OpenDeviceError(
+                            idx.to_string(),
+                            why.to_string(),
+                        ))
+                    }
                 };
                 vid_cap
             }
             CameraIndexType::IPCamera(ip) => match VideoCapture::from_file(&*ip, CAP_ANY) {
                 Ok(vc) => vc,
-                Err(why) => return Err(NokhwaError::CouldntOpenDevice(why.to_string())),
+                Err(why) => return Err(NokhwaError::OpenDeviceError(ip, why.to_string())),
             },
         };
 
@@ -200,9 +214,9 @@ impl OpenCvCaptureDevice {
     /// # Errors
     /// If the frame is failed to be read, this will error.
     #[allow(clippy::cast_sign_loss)]
-    pub fn raw_frame_vec(&mut self) -> Result<Vec<u8>, NokhwaError> {
+    pub fn raw_frame_vec(&mut self) -> Result<Cow<[u8]>, NokhwaError> {
         if !self.is_stream_open() {
-            return Err(NokhwaError::CouldntCaptureFrame(
+            return Err(NokhwaError::ReadFrameError(
                 "Stream is not open!".to_string(),
             ));
         }
@@ -211,13 +225,13 @@ impl OpenCvCaptureDevice {
         match self.video_capture.read(&mut frame) {
             Ok(a) => {
                 if !a {
-                    return Err(NokhwaError::CouldntCaptureFrame(
+                    return Err(NokhwaError::ReadFrameError(
                         "Failed to read frame from videocapture: OpenCV return false, camera disconnected?".to_string(),
                     ));
                 }
             }
             Err(why) => {
-                return Err(NokhwaError::CouldntCaptureFrame(format!(
+                return Err(NokhwaError::ReadFrameError(format!(
                     "Failed to read frame from videocapture: {}",
                     why.to_string()
                 )))
@@ -227,7 +241,7 @@ impl OpenCvCaptureDevice {
         let frame_empty = match frame.empty() {
             Ok(e) => e,
             Err(why) => {
-                return Err(NokhwaError::CouldntCaptureFrame(format!(
+                return Err(NokhwaError::ReadFrameError(format!(
                     "Failed to check for empty OpenCV frame: {}",
                     why.to_string()
                 )))
@@ -246,7 +260,7 @@ impl OpenCvCaptureDevice {
                                 let frame_data_vec = match Mat::data_typed::<Vec3b>(&frame) {
                                     Ok(v) => v,
                                     Err(why) => {
-                                        return Err(NokhwaError::CouldntCaptureFrame(format!(
+                                        return Err(NokhwaError::ReadFrameError(format!(
                                             "Failed to convert frame into raw Vec3b: {}",
                                             why.to_string()
                                         )))
@@ -260,24 +274,24 @@ impl OpenCvCaptureDevice {
                                     raw_vec.push(pixel_slice[0]);
                                 }
 
-                                return Ok(raw_vec);
+                                return Ok(Cow::from(raw_vec));
                             }
-                            Err(NokhwaError::CouldntCaptureFrame(
+                            Err(NokhwaError::ReadFrameError(
                                 "Failed to read frame from videocapture: not cont".to_string(),
                             ))
                         }
-                        Err(why) => Err(NokhwaError::CouldntCaptureFrame(format!(
+                        Err(why) => Err(NokhwaError::ReadFrameError(format!(
                             "Failed to read frame from videocapture: failed to read continuous: {}",
                             why.to_string()
                         ))),
                     };
                 }
-                Err(NokhwaError::CouldntCaptureFrame(
+                Err(NokhwaError::ReadFrameError(
                     "Frame width is less than zero!".to_string(),
                 ))
             }
             Err(why) => {
-                return Err(NokhwaError::CouldntCaptureFrame(format!(
+                return Err(NokhwaError::ReadFrameError(format!(
                     "Failed to read frame from videocapture: failed to read size: {}",
                     why.to_string()
                 )))
@@ -294,7 +308,7 @@ impl OpenCvCaptureDevice {
         let width = match self.video_capture.get(CAP_PROP_FRAME_WIDTH) {
             Ok(width) => width as u32,
             Err(why) => {
-                return Err(NokhwaError::CouldntQueryDevice {
+                return Err(NokhwaError::GetPropertyError {
                     property: "Width".to_string(),
                     error: why.to_string(),
                 })
@@ -304,7 +318,7 @@ impl OpenCvCaptureDevice {
         let height = match self.video_capture.get(CAP_PROP_FRAME_HEIGHT) {
             Ok(height) => height as u32,
             Err(why) => {
-                return Err(NokhwaError::CouldntQueryDevice {
+                return Err(NokhwaError::GetPropertyError {
                     property: "Height".to_string(),
                     error: why.to_string(),
                 })
@@ -322,7 +336,7 @@ impl OpenCvCaptureDevice {
     pub fn raw_framerate(&self) -> Result<u32, NokhwaError> {
         match self.video_capture.get(CAP_PROP_FPS) {
             Ok(fps) => Ok(fps as u32),
-            Err(why) => Err(NokhwaError::CouldntQueryDevice {
+            Err(why) => Err(NokhwaError::GetPropertyError {
                 property: "Framerate".to_string(),
                 error: why.to_string(),
             }),
@@ -331,8 +345,12 @@ impl OpenCvCaptureDevice {
 }
 
 impl CaptureBackendTrait for OpenCvCaptureDevice {
-    fn camera_info(&self) -> CameraInfo {
-        self.camera_info.clone()
+    fn backend(&self) -> CaptureAPIBackend {
+        CaptureAPIBackend::OpenCv
+    }
+
+    fn camera_info(&self) -> &CameraInfo {
+        &self.camera_info
     }
 
     fn camera_format(&self) -> CameraFormat {
@@ -344,7 +362,7 @@ impl CaptureBackendTrait for OpenCvCaptureDevice {
         let is_opened = match self.video_capture.is_opened() {
             Ok(opened) => opened,
             Err(why) => {
-                return Err(NokhwaError::CouldntQueryDevice {
+                return Err(NokhwaError::GetPropertyError {
                     property: "Is Stream Open".to_string(),
                     error: why.to_string(),
                 })
@@ -360,21 +378,28 @@ impl CaptureBackendTrait for OpenCvCaptureDevice {
         if is_opened {
             self.stop_stream()?;
             if let Err(why) = self.open_stream() {
-                return Err(NokhwaError::CouldntOpenDevice(why.to_string()));
+                return Err(NokhwaError::OpenDeviceError(
+                    self.camera_location.to_string(),
+                    why.to_string(),
+                ));
             }
         }
         Ok(())
     }
 
     fn compatible_list_by_resolution(
-        &self,
+        &mut self,
         _fourcc: FrameFormat,
     ) -> Result<HashMap<Resolution, Vec<u32>>, NokhwaError> {
-        Err(NokhwaError::UnsupportedOperation(CaptureAPIBackend::OpenCv))
+        Err(NokhwaError::UnsupportedOperationError(
+            CaptureAPIBackend::OpenCv,
+        ))
     }
 
     fn compatible_fourcc(&mut self) -> Result<Vec<FrameFormat>, NokhwaError> {
-        Err(NokhwaError::UnsupportedOperation(CaptureAPIBackend::OpenCv))
+        Err(NokhwaError::UnsupportedOperationError(
+            CaptureAPIBackend::OpenCv,
+        ))
     }
 
     fn resolution(&self) -> Resolution {
@@ -394,7 +419,7 @@ impl CaptureBackendTrait for OpenCvCaptureDevice {
 
     fn set_frame_rate(&mut self, new_fps: u32) -> Result<(), NokhwaError> {
         let mut current_fmt = self.camera_format;
-        current_fmt.set_framerate(new_fps);
+        current_fmt.set_frame_rate(new_fps);
         self.set_camera_format(current_fmt)
     }
 
@@ -408,6 +433,96 @@ impl CaptureBackendTrait for OpenCvCaptureDevice {
         self.set_camera_format(current_fmt)
     }
 
+    fn supported_camera_controls(&self) -> Result<Vec<KnownCameraControls>, NokhwaError> {
+        Err(NokhwaError::UnsupportedOperationError(
+            CaptureAPIBackend::UniversalVideoClass,
+        ))
+    }
+
+    fn camera_control(&self, _control: KnownCameraControls) -> Result<CameraControl, NokhwaError> {
+        Err(NokhwaError::UnsupportedOperationError(
+            CaptureAPIBackend::UniversalVideoClass,
+        ))
+    }
+
+    fn set_camera_control(&mut self, _control: CameraControl) -> Result<(), NokhwaError> {
+        Err(NokhwaError::UnsupportedOperationError(
+            CaptureAPIBackend::UniversalVideoClass,
+        ))
+    }
+
+    fn raw_supported_camera_controls(&self) -> Result<Vec<Box<dyn Any>>, NokhwaError> {
+        Err(NokhwaError::UnsupportedOperationError(
+            CaptureAPIBackend::UniversalVideoClass,
+        ))
+    }
+
+    fn raw_camera_control(&self, control: &dyn Any) -> Result<Box<dyn Any>, NokhwaError> {
+        let id = match control.downcast_ref::<i32>() {
+            Some(id) => *id,
+            None => {
+                return Err(NokhwaError::StructureError {
+                    structure: "OpenCV ID".to_string(),
+                    error: "Failed Any Cast".to_string(),
+                })
+            }
+        };
+
+        match self.video_capture.get(id) {
+            Ok(v) => Ok(Box::new(v)),
+            Err(why) => {
+                return Err(NokhwaError::GetPropertyError {
+                    property: format!("OpenCV PROP ID {}", id),
+                    error: why.to_string(),
+                })
+            }
+        }
+    }
+
+    fn set_raw_camera_control(
+        &mut self,
+        control: &dyn Any,
+        value: &dyn Any,
+    ) -> Result<(), NokhwaError> {
+        let id = match control.downcast_ref::<i32>() {
+            Some(id) => *id,
+            None => {
+                return Err(NokhwaError::StructureError {
+                    structure: "OpenCV ID".to_string(),
+                    error: "Failed Any Cast".to_string(),
+                })
+            }
+        };
+
+        let value = match value.downcast_ref::<f64>() {
+            Some(id) => *id,
+            None => {
+                return Err(NokhwaError::StructureError {
+                    structure: "OpenCV Value".to_string(),
+                    error: "Failed Any Cast".to_string(),
+                })
+            }
+        };
+
+        match self.video_capture.set(id, value) {
+            Ok(b) => {
+                if b {
+                    return Ok(());
+                }
+                Err(NokhwaError::SetPropertyError {
+                    property: format!("OpenCV PROP ID {}", id),
+                    value: value.to_string(),
+                    error: "OpenCV returned false".to_string(),
+                })
+            }
+            Err(why) => Err(NokhwaError::SetPropertyError {
+                property: format!("OpenCV PROP ID {}", id),
+                value: value.to_string(),
+                error: why.to_string(),
+            }),
+        }
+    }
+
     #[allow(clippy::cast_possible_wrap)]
     fn open_stream(&mut self) -> Result<(), NokhwaError> {
         match self.camera_location.clone() {
@@ -418,10 +533,10 @@ impl CaptureBackendTrait for OpenCvCaptureDevice {
                 {
                     Ok(_) => {}
                     Err(why) => {
-                        return Err(NokhwaError::CouldntOpenDevice(format!(
-                            "Failed to open device: {}",
-                            why.to_string()
-                        )))
+                        return Err(NokhwaError::OpenDeviceError(
+                            idx.to_string(),
+                            format!("Failed to open device: {}", why.to_string()),
+                        ))
                     }
                 }
             }
@@ -432,10 +547,10 @@ impl CaptureBackendTrait for OpenCvCaptureDevice {
                 {
                     Ok(_) => {}
                     Err(why) => {
-                        return Err(NokhwaError::CouldntOpenDevice(format!(
-                            "Failed to open device: {}",
-                            why.to_string()
-                        )))
+                        return Err(NokhwaError::OpenDeviceError(
+                            ip,
+                            format!("Failed to open device: {}", why.to_string()),
+                        ))
                     }
                 }
             }
@@ -446,11 +561,11 @@ impl CaptureBackendTrait for OpenCvCaptureDevice {
                 if open {
                     return Ok(());
                 }
-                Err(NokhwaError::CouldntOpenStream(
+                Err(NokhwaError::OpenStreamError(
                     "Stream is not opened after stream open attempt opencv".to_string(),
                 ))
             }
-            Err(why) => Err(NokhwaError::CouldntQueryDevice {
+            Err(why) => Err(NokhwaError::GetPropertyError {
                 property: "Is Stream Open After Open Stream".to_string(),
                 error: why.to_string(),
             }),
@@ -462,13 +577,16 @@ impl CaptureBackendTrait for OpenCvCaptureDevice {
     }
 
     fn frame(&mut self) -> Result<ImageBuffer<Rgb<u8>, Vec<u8>>, NokhwaError> {
-        let mut image_data = self.frame_raw()?;
         let camera_resolution = self.camera_format.resolution();
-        image_data.resize(
-            (camera_resolution.width() * camera_resolution.height() * 3) as usize,
-            0_u8,
-        );
-        let imagebuf =
+        let image_data = {
+            let mut data = self.frame_raw()?.to_vec();
+            data.resize(
+                (camera_resolution.width() * camera_resolution.height() * 3) as usize,
+                0_u8,
+            );
+            data
+        };
+        let image_buf =
             match ImageBuffer::from_vec(
                 camera_resolution.width(),
                 camera_resolution.height(),
@@ -478,24 +596,23 @@ impl CaptureBackendTrait for OpenCvCaptureDevice {
                     let rgb: ImageBuffer<Rgb<u8>, Vec<u8>> = buf;
                     rgb
                 }
-                None => return Err(NokhwaError::CouldntCaptureFrame(
+                None => return Err(NokhwaError::ReadFrameError(
                     "Imagebuffer is not large enough! This is probably a bug, please report it!"
                         .to_string(),
                 )),
             };
-        Ok(imagebuf)
+        Ok(image_buf)
     }
 
-    fn frame_raw(&mut self) -> Result<Vec<u8>, NokhwaError> {
-        let vec = self.raw_frame_vec()?;
-        // println!("{:?}", vec);
-        Ok(vec)
+    fn frame_raw(&mut self) -> Result<Cow<[u8]>, NokhwaError> {
+        let cow = self.raw_frame_vec()?;
+        Ok(cow)
     }
 
     fn stop_stream(&mut self) -> Result<(), NokhwaError> {
         match self.video_capture.release() {
             Ok(_) => Ok(()),
-            Err(why) => Err(NokhwaError::CouldntStopStream(why.to_string())),
+            Err(why) => Err(NokhwaError::StreamShutdownError(why.to_string())),
         }
     }
 }
